/**
 * Sencha GXT 4.0.0 - Sencha for GWT
 * Copyright (c) 2006-2015, Sencha Inc.
 *
 * licensing@sencha.com
 * http://www.sencha.com/products/gxt/license/
 *
 * ================================================================================
 * Open Source License
 * ================================================================================
 * This version of Sencha GXT is licensed under the terms of the Open Source GPL v3
 * license. You may use this license only if you are prepared to distribute and
 * share the source code of your application under the GPL v3 license:
 * http://www.gnu.org/licenses/gpl.html
 *
 * If you are NOT prepared to distribute and share the source code of your
 * application under the GPL v3 license, other commercial and oem licenses
 * are available for an alternate download of Sencha GXT.
 *
 * Please see the Sencha GXT Licensing page at:
 * http://www.sencha.com/products/gxt/license/
 *
 * For clarification or additional options, please contact:
 * licensing@sencha.com
 * ================================================================================
 *
 *
 * ================================================================================
 * Disclaimer
 * ================================================================================
 * THIS SOFTWARE IS DISTRIBUTED "AS-IS" WITHOUT ANY WARRANTIES, CONDITIONS AND
 * REPRESENTATIONS WHETHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
 * IMPLIED WARRANTIES AND CONDITIONS OF MERCHANTABILITY, MERCHANTABLE QUALITY,
 * FITNESS FOR A PARTICULAR PURPOSE, DURABILITY, NON-INFRINGEMENT, PERFORMANCE AND
 * THOSE ARISING BY STATUTE OR FROM CUSTOM OR USAGE OF TRADE OR COURSE OF DEALING.
 * ================================================================================
 */
package com.sencha.gxt.themebuilder.config.shared;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import com.sencha.gxt.themebuilder.config.shared.AbstractCfgObject.CfgVisitor;

public class PrintVisitor extends CfgVisitor {
  private final StringBuilder sb = new StringBuilder();
  private int depth = 0;
  private boolean newLine = true;

  @Override
  public void visitDouble(CfgDoubleLiteral cfg) {
    append(cfg.resolveDouble());
  }

  @Override
  public void visitString(CfgStringLiteral cfg) {
    append("\"").append(escape(cfg.resolveString())).append("\"");
  }

  @Override
  public void visitInt(CfgIntLiteral cfg) {
    append(cfg.resolveInt());
  }

  @Override
  public void visitObject(CfgObject cfg) {
    //sort keys to get leaf nodes first, then get objects
    List<String> leafNodes = new ArrayList<String>();
    List<String> objects = new ArrayList<String>();

    for (String key : cfg.keySet()) {
      AbstractCfgObject value = cfg.get(key);
      if (value instanceof CfgObject || (value instanceof CfgReference && (((CfgReference) value).tryResolve() instanceof CfgObject))) {
        objects.add(key);
      } else {
        leafNodes.add(key);
      }
    }
    //sort each list alphabetically
    Collections.sort(leafNodes);
    Collections.sort(objects);

    List<String> keys = leafNodes;
    keys.addAll(objects);

    for (String key : keys) {
      if (cfg.getRemark(key) != null) {
        append("// ").append(cfg.getRemark(key));
        println();
      }

      append(key);

      AbstractCfgObject obj = cfg.get(key);
      if (obj instanceof CfgObject) {
        append(" {");
        indent();
        println();
        obj.accept(this);
        outdent();
        append("}");
      } else {
        append(" = ");
        obj.accept(this);
      }
      println();
    }
  }

  @Override
  public void visitArray(CfgArray cfg) {
    append("[");
    indent();
    println();

    boolean first = true;
    for (AbstractCfgObject obj : cfg.getItems()) {
      if (!first) {
        append(",");
        println();
      }
      first = false;
      obj.accept(this);
    }

    outdent();
    println();
    append("]");
  }


  @Override
  public void visitReference(CfgReference cfg) {
    append(cfg.getReference());
  }

  @Override
  public void visitExpression(CfgExpression cfg) {
    cfg.getLhs().accept(this);

    append(" ");
    switch (cfg.getOp()) {
      case PLUS:
        append("+");
        break;
      case MINUS:
        append("-");
        break;
      case MULTIPLY:
        append("*");
        break;
      case DIVIDE:
        append("/");
        break;
    }
    append(" ");

    cfg.getRhs().accept(this);
  }

  @Override
  public void visitFuncCall(CfgFuncCall cfg) {
    append("util.").append(cfg.getMethod()).append("(");
    boolean first = true;
    for (AbstractCfgObject arg : cfg.getArgs()) {
      if (!first) {
        append(", ");
      }
      first = false;
      if (arg instanceof CfgObject){
        // if an argument is a CfgObject, traverse appropriately with scope block
        append("{");
        println();
        indent();
        arg.accept(this);
        outdent();
        append("}");
      } else {
        arg.accept(this);
      }
    }
    append(")");
  }

  @Override
  public String toString() {
    return sb.toString();
  }

  private void indent() {
    depth++;
  }
  private void outdent() {
    assert depth > 0;
    depth--;
  }
  private void println() {
    assert !newLine : "Double \\n";
    sb.append("\n");
    newLine = true;
  }
  private StringBuilder append(int i) {
    return append("" + i);
  }
  private StringBuilder append(double d) {
    return append("" + d);
  }
  private StringBuilder append(String str) {
    if (newLine) {
      doIndent();
    }
    newLine = false;
    return sb.append(str);
  }

  private void doIndent() {
    for (int i = 0; i < depth; i++) {
      sb.append("  ");
    }
  }

  /**
   * Escapes string content to be a valid string literal.
   *
   * @return an escaped version of <code>unescaped</code>, suitable for being
   *         enclosed in double quotes in Java source
   */
  public static String escape(String unescaped) {
    int extra = 0;
    int i = 0;
    int n = unescaped.length();
    for (; i < n; ++i) {
      switch (unescaped.charAt(i)) {
        case '\0':
        case '\n':
        case '\r':
        case '\"':
        case '\\':
          ++extra;
          break;
      }
    }

    if (extra == 0) {
      return unescaped;
    }

    char[] oldChars = unescaped.toCharArray();
    char[] newChars = new char[oldChars.length + extra];
    i = 0;
    int out = 0;
    n = oldChars.length;
    for (; i < n; ++i) {
      char c = oldChars[i];
      switch (c) {
        case '\0':
          newChars[out++] = '\\';
          c = '0';
          break;
        case '\n':
          newChars[out++] = '\\';
          c = 'n';
          break;
        case '\r':
          newChars[out++] = '\\';
          c = 'r';
          break;
        case '\"':
          newChars[out++] = '\\';
          c = '"';
          break;
        case '\\':
          newChars[out++] = '\\';
          c = '\\';
          break;
      }
      newChars[out] = c;
      ++out;
    }

    return String.valueOf(newChars);
  }
}
