/**
 * Sencha GXT 4.0.0 - Sencha for GWT
 * Copyright (c) 2006-2015, Sencha Inc.
 *
 * licensing@sencha.com
 * http://www.sencha.com/products/gxt/license/
 *
 * ================================================================================
 * Open Source License
 * ================================================================================
 * This version of Sencha GXT is licensed under the terms of the Open Source GPL v3
 * license. You may use this license only if you are prepared to distribute and
 * share the source code of your application under the GPL v3 license:
 * http://www.gnu.org/licenses/gpl.html
 *
 * If you are NOT prepared to distribute and share the source code of your
 * application under the GPL v3 license, other commercial and oem licenses
 * are available for an alternate download of Sencha GXT.
 *
 * Please see the Sencha GXT Licensing page at:
 * http://www.sencha.com/products/gxt/license/
 *
 * For clarification or additional options, please contact:
 * licensing@sencha.com
 * ================================================================================
 *
 *
 * ================================================================================
 * Disclaimer
 * ================================================================================
 * THIS SOFTWARE IS DISTRIBUTED "AS-IS" WITHOUT ANY WARRANTIES, CONDITIONS AND
 * REPRESENTATIONS WHETHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE
 * IMPLIED WARRANTIES AND CONDITIONS OF MERCHANTABILITY, MERCHANTABLE QUALITY,
 * FITNESS FOR A PARTICULAR PURPOSE, DURABILITY, NON-INFRINGEMENT, PERFORMANCE AND
 * THOSE ARISING BY STATUTE OR FROM CUSTOM OR USAGE OF TRADE OR COURSE OF DEALING.
 * ================================================================================
 */
package com.sencha.gxt.themebuilder.config.shared;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.util.ArrayList;
import java.util.List;

import com.sencha.gxt.themebuilder.graphics.shared.Color;
import com.sencha.gxt.themebuilder.graphics.shared.Css3GradientInfo;
import com.sencha.gxt.themebuilder.graphics.shared.GraphicsUtil;

public class ThemerUtils {
  public static ThemerUtilFactory factory;
  @Retention(RetentionPolicy.RUNTIME)
  @Target(ElementType.METHOD)
  public @interface UtilMethod {
  }
  @Retention(RetentionPolicy.RUNTIME)
  @Target(ElementType.PARAMETER)
  public @interface Optional {
    String value();
  }

  public enum UtilMethodType {
    INT {
      @Override
      public Object value(AbstractCfgObject obj) {
        return obj.resolveInt();
      }
    },
    DOUBLE {
      @Override
      public Object value(AbstractCfgObject obj) {
        return obj.resolveDouble();
      }
    },
    STRING {
      @Override
      public Object value(AbstractCfgObject obj) {
        return obj.resolveString();
      }
    },
    OBJECT{
      @Override
      public Object value(AbstractCfgObject obj) {
        return obj.resolveObject();
      }
    };
    public abstract Object value(AbstractCfgObject obj);
  }

  public static abstract class ThemerUtilFactory {
    public ThemerUtilInvocation invokeWith(SourceInfo source, CfgObject scope) {
      return new ThemerUtilInvocation(source, scope);
    }
    public class ThemerUtilInvocation {
      private final SourceInfo source;
      private final CfgObject scope;
      public ThemerUtilInvocation(SourceInfo source, CfgObject scope) {
        this.source = source;
        this.scope = scope;
      }
      public String stringMethod(String name, List<AbstractCfgObject> args) {
        return (String) method(name, args, UtilMethodType.STRING);
      }
      public double doubleMethod(String name, List<AbstractCfgObject> args) {
        return (Double) method(name, args, UtilMethodType.DOUBLE);
      }
      public int intMethod(String name, List<AbstractCfgObject> args) {
        return (Integer) method(name, args, UtilMethodType.INT);
      }
      public CfgObject objectMethod(String name, List<AbstractCfgObject> args) {
        return (CfgObject) method(name, args, UtilMethodType.OBJECT);
      }

      public Object method(String name, List<AbstractCfgObject> args, UtilMethodType returnType) {
        ThemeMethod method = getMethod(new ThemerUtils(source, scope), name, returnType, args.size());
        return method.invoke(getArgs(method.argType(), args));
      }
    }

    private Object[] getArgs(UtilMethodType[] types, List<AbstractCfgObject> args) {
      assert types.length == args.size();
      Object[] ret = new Object[args.size()];
      for (int i = 0; i < args.size(); i++) {
        ret[i] = types[i].value(args.get(i));
      }
      return ret;
    }

    public abstract ThemeMethod getMethod(ThemerUtils utils, String name, UtilMethodType returnValue, int count);
    public interface ThemeMethod {
      Object invoke(Object... args);
      UtilMethodType[] argType();
    }
  }

  private final SourceInfo source;
  private final CfgObject scope;
  public ThemerUtils(SourceInfo source, CfgObject scope) {
    this.source = source;
    this.scope = scope;
  }

  /**
   * Darkens the given gradient data structure by the specified amount
   *
   * @param gradient the gradient data
   * @param scale
   * @return a new gradient object
   */
  @UtilMethod
  public CfgObject darkenGradient(CfgObject gradient, double scale) {
    CfgObject darkGradient = new CfgObject(source, scope);
    List<AbstractCfgObject> stops = new ArrayList<AbstractCfgObject>();
    for (AbstractCfgObject stop : gradient.get("stops").resolveArray().getItems()) {
      Css3GradientInfo grad = Css3GradientInfo.valueOf(stop.resolveString());
      String colorChange = GraphicsUtil.colorToHexString(GraphicsUtil.darken(grad.getColor(), scale));
      String newValue = colorChange + " " + grad.getPercentString();
      stops.add(new CfgStringLiteral(source, newValue));
    }
    CfgArray array = new CfgArray(source, stops);
    darkGradient.put("stops", array);
    return darkGradient;
  }


  @UtilMethod
  public CfgObject lightenGradient(CfgObject gradient, double scale) {
    CfgObject lightGradient = new CfgObject(source, scope);
    List<AbstractCfgObject> stops = new ArrayList<AbstractCfgObject>();
    for (AbstractCfgObject stop : gradient.get("stops").resolveArray().getItems()) {
      Css3GradientInfo grad = Css3GradientInfo.valueOf(stop.resolveString());
      String colorChange = GraphicsUtil.colorToHexString(GraphicsUtil.lighten(grad.getColor(), scale));
      String newValue = colorChange + " " + grad.getPercentString();
      stops.add(new CfgStringLiteral(source, newValue));
    }
    CfgArray array = new CfgArray(source, stops);
    lightGradient.put("stops", array);
    return lightGradient;
  }

  @UtilMethod
  public CfgObject solidGradient(String colorString) {
    CfgObject gradient = new CfgObject(source, scope);
    ArrayList<AbstractCfgObject> stops = new ArrayList<AbstractCfgObject>();
    stops.add(new CfgStringLiteral(source, colorString + " 0%"));
    stops.add(new CfgStringLiteral(source, colorString + " 100%"));
    CfgArray array = new CfgArray(source, stops);
    gradient.put("stops", array);
    return gradient;
  }

  @UtilMethod
  public String solidGradientString(String colorString) {
    return colorString + " 0%, " + colorString + " 100%";
  }

  @UtilMethod
  public String darkenColor(String colorString, double scale) {
    return GraphicsUtil.colorToHexString(GraphicsUtil.darken(Color.decode(colorString), scale));
  }
  @UtilMethod
  public String lightenColor(String colorString, double scale) {
    return GraphicsUtil.colorToHexString(GraphicsUtil.lighten(Color.decode(colorString), scale));
  }

  @UtilMethod
  public String mixColors(String color1, String color2) {
    return GraphicsUtil.colorToHexString(GraphicsUtil.mix(Color.decode(color1), Color.decode(color2)));
  }
  @UtilMethod
  public String mixColors(String color1, String color2, double ratio) {
    return GraphicsUtil.colorToHexString(GraphicsUtil.mix(Color.decode(color1), Color.decode(color2), ratio));
  }

  @UtilMethod
  public String gradientString(CfgObject gradient) {
    StringBuilder sb = new StringBuilder();

    boolean first = true;
    for (AbstractCfgObject grad : gradient.get("stops").resolveArray().getItems()) {
      if (!first) {
        sb.append(", ");
      }
      first = false;
      sb.append(grad.resolveString());
    }
    return sb.toString();
  }
  @UtilMethod
  public String lastStopColor(CfgObject gradient) {
    List<AbstractCfgObject> stops = gradient.get("stops").resolveArray().getItems();
    return GraphicsUtil.colorToHexString(Css3GradientInfo.valueOf(stops.get(stops.size() - 1).resolveString()).getColor());
  }
  @UtilMethod
  public String colorNameToHex(String name) {
    return GraphicsUtil.colorNameToHexString(name);
  }

  @UtilMethod
  public String rgbToHex(int r, int g, int b) {
    return GraphicsUtil.rgbToHexString(r, g, b);
  }

  @UtilMethod
  public CfgObject margin(int top, @Optional("top")int right, @Optional("top")int bottom, @Optional("side2")int left) {
    return edge(top, right, bottom, left);
  }
  @UtilMethod
  public CfgObject margin(int top, int right, int bottom) {
    return edge(top, right, bottom);
  }
  @UtilMethod
     public CfgObject margin(int top, int right) {
    return edge(top, right);
  }
  @UtilMethod
  public CfgObject margin(int top) {
    return edge(top);
  }
  @UtilMethod
  public CfgObject padding(int top, @Optional("top")int right, @Optional("top")int bottom, @Optional("side2")int left) {
    return edge(top, right, bottom, left);
  }
  @UtilMethod
  public CfgObject padding(int side1, int side2, int side3) {
    return edge(side1, side2, side3);
  }
  @UtilMethod
  public CfgObject padding(int side1, int side2) {
    return edge(side1, side2);
  }
  @UtilMethod
  public CfgObject padding(int side1) {
    return edge(side1);
  }

  public CfgObject edge(int... sides) {
    CfgObject obj = new CfgObject(source, scope);

    obj.put("top", sides[0]);
    obj.put("right", sides.length > 1 ? sides[1] : sides[0]);
    obj.put("bottom", sides.length > 2 ? sides[2] : obj.get("top").resolveInt());
    obj.put("left", sides.length > 3 ? sides[3] : obj.get("right").resolveInt());

    return obj;
  }
  public CfgObject edge(String... sides) {
    CfgObject obj = new CfgObject(source, scope);

    obj.put("top", sides[0]);
    obj.put("right", sides.length > 1 ? sides[1] : sides[0]);
    obj.put("bottom", sides.length > 2 ? sides[2] : obj.get("top").resolveString());
    obj.put("left", sides.length > 3 ? sides[3] : obj.get("right").resolveString());

    return obj;
  }

  @UtilMethod
  public CfgObject border(String style) {
    return borderImpl(style, "#000000", 0);
  }
  @UtilMethod
  public CfgObject border(String style, String color) {
    return borderImpl(style, color, 0);
  }
  @UtilMethod
  public CfgObject border(String style, String color, int top) {
    return borderImpl(style, color, top);
  }
  @UtilMethod
  public CfgObject border(String style, String color, int top, int right) {
    return borderImpl(style, color, top, right);
  }
  @UtilMethod
  public CfgObject border(String style, String color, int top, int right, int bottom) {
    return borderImpl(style, color, top, right, bottom);
  }
  @UtilMethod
  public CfgObject border(String style, String color, int top, @Optional("top")int right, @Optional("top")int bottom, @Optional("right")int left) {
    return borderImpl(style, color, top, right, bottom, left);
  }
  public CfgObject borderImpl(String style, String color, int... sides) {
  CfgObject config = edge(sides);

    config.put("style", style);
    config.put("color", color);

    return config;
  }
  @UtilMethod
  public int abs(int value) {
    return Math.abs(value);
  }
  @UtilMethod
  public double abs(double value) {
    return Math.abs(value);
  }
  @UtilMethod
  public int min(int a, int b) {
    return Math.min(a, b);
  }
  @UtilMethod
  public double min(double a, double b) {
    return Math.min(a, b);
  }
  @UtilMethod
  public int max(int a, int b) {
    return Math.max(a, b);
  }
  @UtilMethod
  public double max(double a, double b) {
    return Math.max(a, b);
  }

  @UtilMethod
  public CfgObject extend(CfgObject originalObject, CfgObject additionalProperties) {
    assert originalObject != additionalProperties;
    CfgObject obj = new CfgObject(source, scope);

    //clone additionalProperties, then if originalObject has something which additionalProperties doesn't, copy it in
    internalCopy(additionalProperties, obj);//clone
    internalCopy(originalObject, obj);//overwrite missing

    return obj;
  }

  private void internalCopy(CfgObject from, CfgObject to) {
    for (String key : from.keySet()) {
      AbstractCfgObject value = from.get(key);

      if (to.keySet().contains(key)) {
        //first check if we are looking at a obj so we should descend and copy more
        if (value instanceof CfgObject) {
          CfgObject subValue = new CfgObject(source, to);
          internalCopy(to.get(key).resolveObject(), subValue);//clone
          internalCopy((CfgObject) value, subValue);//overwrite missing
          value = subValue;
          to.erase(key);
        } else {
          //next, if not an obj already and to has, skip it
          continue;
        }
      }

      //clone the rest of the object to rescope it to its new parent
      to.put(key, cloneValue(to, value));
    }
  }

  private AbstractCfgObject cloneValue(CfgObject to, final AbstractCfgObject value) {
    if (value instanceof CfgReference) {
      CfgReference ref = (CfgReference) value;
      return new CfgReference(ref.getSource(), to, ref.getReference(), ref.getName());
    } else if (value instanceof CfgExpression) {
      CfgExpression expr = (CfgExpression) value;
      return new CfgExpression(expr.getSource(), cloneValue(to, expr.getLhs()), cloneValue(to, expr.getRhs()), expr.getOp());
    } else if (value instanceof CfgFuncCall) {
      CfgFuncCall call = (CfgFuncCall) value;
      List<AbstractCfgObject> newArgs = new ArrayList<AbstractCfgObject>(call.getArgs().size());
      for (AbstractCfgObject a : call.getArgs()) {
        newArgs.add(cloneValue(to, a));
      }
      return new CfgFuncCall(call.getSource(), to, call.getMethod(), newArgs);
    } else if (value instanceof CfgObject) {
      CfgObject obj = (CfgObject) value;
      //unreferencable, go ahead and re-root recurively, no need to merge
      CfgObject newObj = new CfgObject(obj.getSource(), to);
      internalCopy(obj, newObj);
      return newObj;
    }
    return value;
  }

  @UtilMethod
  public CfgObject fontStyle(String family) {
    return fontStyle(family, "medium");
  }
  @UtilMethod
  public CfgObject fontStyle(String family, String size) {
    return fontStyle(family, size, "#000000");
  }
  @UtilMethod
  public CfgObject fontStyle(String family, String size, String color) {
    return fontStyle(family, size, color, "normal");
  }
  @UtilMethod
  public CfgObject fontStyle(String family, @Optional("'medium'")String size, @Optional("'#000000'")String color, @Optional("'normal'")String weight) {
    CfgObject obj = new CfgObject(source, scope);
    obj.put("family", family);
    obj.put("size", size);
    obj.put("color", color);
    obj.put("weight", weight);
    return obj;
  }

  @UtilMethod
  public CfgObject radiusMinusBorderWidth(CfgObject borderObject, int radius) {
    CfgObject obj = new CfgObject(source, scope);

    obj.put("top", Math.max(0, radius - borderObject.get("top").resolveInt()));
    obj.put("right", Math.max(0, radius - borderObject.get("right").resolveInt()));
    obj.put("bottom", Math.max(0, radius - borderObject.get("bottom").resolveInt()));
    obj.put("left", Math.max(0, radius - borderObject.get("left").resolveInt()));

    return obj;
  }

  /*
   * Each of these functions has a default 'scale' value the inner values are based on that. If that scale is set to 1,
   * then the function will max out in one direction (or possibly both) at white or black. Any given function has two or
   * three numeric values that is adjusts the base color by, call them a, b, c. The default scale of a function is then
   * defaultScale = max(a, b, c), and the other values are adjusted a' = a / defaultScale * scale. When the scale is the
   * default, then a' == a, etc. Likewise at scale = 1, since defaultScale / defaultScale == 1, the max of a, b, c will
   * be 1, and the other values will be at their various extremes. For further stretching (i.e. all the way from black
   * to white), the input color will need to be pre-darkened or -lightened.
   *
   * The original values (both defaultScale and the various a, b, c) are left in as constants for readability.
   */
  @UtilMethod
  public String glossyGradient(String base) {
    return glossyGradient(base, 0.15);
  }
  @UtilMethod
  public String glossyGradient(String base, double scale) {
    scale = Math.max(0.0, Math.min(1.0, scale));
    return stops(
            lightenColor(base, 0.15 / 0.15 * scale),
            lightenColor(base, 0.05 / 0.15 * scale) + " 50%",
            base + " 51%",
            darkenColor(base, 0.05 / 0.15 * scale)
    );
  }

  @UtilMethod
  public String bevelGradient(String base) {
    return bevelGradient(base, 0.15);
  }
  @UtilMethod
  public String bevelGradient(String base, double scale) {
    scale = Math.max(0.0, Math.min(1.0, scale));
    return stops(
            lightenColor(base, 0.15 / 0.15 * scale),
            lightenColor(base, 0.08 / 0.15 * scale) + " 30%",
            base + " 65%",
            darkenColor(base, 0.06 / 0.15 * scale)
    );
  }

  @UtilMethod
  public String recessedGradient(String base) {
    return recessedGradient(base, 0.5);
  }
  @UtilMethod
  public String recessedGradient(String base, double scale) {
    scale = Math.max(0.0, Math.min(1.0, scale));
    return stops(
            darkenColor(base, 0.1 / 0.5 * scale),
            darkenColor(base, 0.05 / 0.5 * scale) + " 10%",
            base + " 51%",
            lightenColor(base, 0.5 / 0.5 * scale)
    );
  }

  @UtilMethod
  public String matteGradient(String base) {
    return matteGradient(base, 0.04);
  }
  @UtilMethod
  public String matteGradient(String base, double scale) {
    scale = Math.max(0.0, Math.min(1.0, scale));
    return stops(
            lightenColor(base, 0.03 / 0.04 * scale),
            darkenColor(base, 0.04 / 0.04 * scale)
    );
  }

  @UtilMethod
  public String reversedMatteGradient(String base) {
    return reversedMatteGradient(base, 0.06);
  }
  @UtilMethod
  public String reversedMatteGradient(String base, double scale) {
    scale = Math.max(0.0, Math.min(1.0, scale));
    return stops(
            darkenColor(base, 0.06 / 0.06 * scale),
            lightenColor(base, 0.04 / 0.06 * scale)
    );
  }

  private String stops(String... stops) {
    CfgObject obj = new CfgObject(source, scope);
    List<AbstractCfgObject> stopsArray = new ArrayList<AbstractCfgObject>();

    for (String stop : stops) {
      stopsArray.add(new CfgStringLiteral(source, stop));
    }

    obj.put("stops", new CfgArray(source, stopsArray));

    return gradientString(obj);
  }
}
